# encoding: utf-8
import os
import re

from argparse import ArgumentParser
from optimize_images.constants import DEFAULT_QUALITY, SUPPORTED_FORMATS


def get_args():
    desc = "A command-line utility written in pure Python to reduce the file " \
           "size of images. You must explicitly pass it a path to the image " \
           "file or to the directory containing the image files to be " \
           "processed."
    epilog = "PLEASE NOTE: The operation is done DESTRUCTIVELY, " \
             "by replacing the original files with the processed ones. You " \
             "definitely should duplicate the original file or folder before " \
             "using this utility, in order to be able to recover any damaged " \
             "images that don't have the desired quality. When doing format " \
             "conversion, if a JPEG with the same name already exists, it " \
             "be replaced by the JPEG file resulting from that conversion."
    parser = ArgumentParser(description=desc, epilog=epilog)

    path_help = "The path to the image file or to the folder containing the " \
                "images to be optimized. By default, it will try to process " \
                "any images found in all of its subdirectories."
    parser.add_argument('path', nargs="?", type=str, help=path_help)

    parser.add_argument('-v', '--version', action='version',
                        version=__import__('optimize_images').__version__)

    sf_help = "Display the list of image formats currently supported."
    parser.add_argument('-s', '--supported', dest="supported_formats",
                        action='store_true', help=sf_help)

    parser.add_argument('-nr', '--no-recursion', action='store_true',
                        help="Don't recurse through subdirectories.")


    general_msg = "These options will be applied individually to each " \
               "image being processed, independently of its format."
    general_group = parser.add_argument_group(
        'General image settings'.upper(), description=general_msg)

    mw_help = "The maximum width (in pixels)."
    general_group.add_argument('-mw', dest="max_width",
                            type=int, default=0, help=mw_help)

    mh_help = "The maximum height (in pixels). Any image that has a dimension " \
               "exceeding a specified value will be downsized as the first " \
               "optimization step. The resizing will not take effect if, " \
               "after the whole optimization process, the resulting file " \
               "size isn't any smaller than the original."
    general_group.add_argument('-mh', dest="max_height",
                            type=int, default=0, help=mh_help)

    general_group.add_argument('-g', '--grayscale', action='store_true',
                        help="Convert to grayscale.")

    nc_help = "Don't compare the original and resulting file sizes, and save " \
             "the new image anyway (useful, for instance, if you prefer to " \
             "have all images with the same color, size, or quality settings)."
    general_group.add_argument('-nc', '--no-comparison', action='store_true',
                        help=nc_help)

    fm_help = "Skip some actions (e.g., the final palete rebuild for indexed " \
              "PNG images or variable JPEG quality setting) in order to " \
              "finish faster."
    general_group.add_argument('-fm', '--fast-mode', action='store_true', help=fm_help)



    jpg_msg = 'The following options apply only to JPEG image files.'
    jpg_group = parser.add_argument_group(
        'JPEG specific options'.upper(), description=jpg_msg)

    q_help = "Specify a fixed quality setting for JPEG files (an integer " \
             "value, between 1 and 100)."
    jpg_group.add_argument('-q', dest="quality",
                           type=int, help=q_help)

    jpg_group.add_argument(
        '-ke',
        "--keep-exif",
        action='store_true',
        help="Keep image EXIF data (by default, it's discarded).")


    png_msg = 'The following options apply only to PNG image files.'
    png_group = parser.add_argument_group(
        'PNG specific options'.upper(), description=png_msg)

    rc_help = "Reduce colors using an adaptive color palette. This option " \
              "can have a big impact both on file size and image quality."
    png_group.add_argument('-rc', "--reduce-colors", dest="reduce_colors",
                           action='store_true', help=rc_help)

    mc_help = "The maximum number of colors when reducing colors (-rc) " \
              "(an integer between 0 and 255). Defaults to 255."
    png_group.add_argument('-mc', dest="max_colors",
                           type=int, default=256, help=mc_help)

    rt_help = "Remove transparency (by default, white background)."
    png_group.add_argument('-rt',
                           dest="remove_transparency", action='store_true',
                           help=rt_help)

    bg_help = "The background color to apply when removing transparency or " \
              "converting to JPEG. Specify 3 integer values (Red, Green and " \
              "Blue), between 0 and 255, separated by spaces. E.g.: " \
              "'255 0 0' for red)."
    png_group.add_argument(
        '-bg', dest="val", type=int, nargs=3, help=bg_help)

    hbg_help = "The background color in hexadecimal (HTML style) to use " \
               "when removing transparency or converting to JPEG. E.g.: '00FF00' " \
               "for green color."
    png_group.add_argument('-hbg', dest="hex_color", type=str, help=hbg_help)

    cb_help = "Convert to JPEG any big PNG images that have " \
              "a large number of colors. It uses an algorithm " \
              "to determine whether it is a good idea and automatically decide " \
              "about it. By default, the original PNG " \
              "files will remain untouched and will be kept alongside the " \
              "optimized JPG images in their original folders (existing JPEGs " \
              "will be replaced)."
    png_group.add_argument(
        '-cb', "--convert-big", action='store_true', help=cb_help)


    ca_help = "Convert to JPEG all PNG images found. By default, " \
              "the original PNG " \
              "files will remain untouched and will be kept alongside the " \
              "optimized JPG images in their original folders (existing JPEGs " \
              "will be replaced)."
    png_group.add_argument(
        '-ca', "--convert-all", action='store_true', help=ca_help)


    fd_help = "Delete the original file when converting to JPG."
    png_group.add_argument(
        '-fd', "--force-delete", action='store_true', help=fd_help)

    parser._positionals.title = parser._positionals.title.upper()
    parser._optionals.title = parser._optionals.title.upper()

    args = parser.parse_args()
    recursive = not args.no_recursion
    quality = args.quality

    if args.supported_formats:
        formats = ', '.join(SUPPORTED_FORMATS).strip().upper()
        msg = "These are the image formats currently supported (please " \
              "note that any files without one of these file extensions " \
              "will be ignored):"
        msg = f"\n{msg} {formats}\n\n"
        parser.exit(status=0, message=msg)

    if args.path:
        src_path = os.path.expanduser(args.path)
    else:
        msg = "\nPlease specify the path of the image or folder to process.\n\n"
        parser.exit(status=0, message=msg)

    if not quality:
        quality = DEFAULT_QUALITY
    elif quality > 100 or quality < 1:
        msg = "\nPlease specify an integer quality value between 1 and 100.\n\n"
        parser.exit(status=0, message=msg)



    if args.max_width < 0 or args.max_height < 0:
        msg = "\nPlease specify image dimensions as positive integers.\n\n"
        parser.exit(status=0, message=msg)

    if args.val and args.hex_color:
        msg = "\nBackground color should be entered only once.\n\n"
        parser.exit(status=0, message=msg)
    elif not args.val and not args.hex_color:
        # By default, apply a white background
        bg_color = (255, 255, 255)
    elif args.val:
        bg_color = tuple(args.val)
    else:
        # Check if hexadecimal is in the expected format
        if not re.search(r'(?:[0-9a-fA-F]{3}){1,2}$', args.hex_color):
            msg = "\nHexadecimal background color was not entered in the correct " \
                  "format. Please follow these examples:\n\nWhite: FFFFFF" \
                  "\nBlack: 000000\nPure Red: FF0000\n\n"
            parser.exit(status=0, message=msg)
        # convert hex to a tuple of integers (RGB)
        bg_color = tuple(
            int(args.hex_color[i:i + 2], 16) for i in (0, 2, 4))

    if min(bg_color) < 0 or max(bg_color) > 255:
        msg = "\nBackground color should be entered as a sequence of 3 " \
              "integer numbers between 0 and 255 (values for Red, Green and " \
              "Blue components) separated by spaces. For instance, for a " \
              "bright red you can use: '-bg 255 0 0' or '-hbg #FF0000'.\n\n"
        parser.exit(status=0, message=msg)

    return src_path, recursive, quality, args.remove_transparency, args.reduce_colors, args.max_colors, \
           args.max_width, args.max_height, args.keep_exif, args.convert_all, args.convert_big, \
           args.force_delete, bg_color, args.grayscale, args.no_comparison, args.fast_mode



# encoding: utf-8
import concurrent.futures
import platform
import shutil
from typing import Tuple, Union

from optimize_images.constants import IOS_FONT, IPHONE_FONT_SIZE, IPAD_FONT_SIZE
from optimize_images.constants import IOS_WORKERS
from optimize_images.data_structures import PPoolExType, TPoolExType


class IconGenerator:
    def __init__(self):
        try:
            if platform.system() in ('Windows', 'Haiku'):
                raise Exception

            print('\n\nUsing these symbols:\n\n'
                  '  ✅ Optimized file     ℹ️  EXIF info present\n'
                  '  🔴 Skipped file       ⤵  Image was downsized     🔻 Size reduction (%)\n')
            self.info = 'ℹ️ '
            self.downsized = '⤵ '
            self.optimized = '✅'
            self.skipped = '🔴'
            self.size_is_smaller = '🔻'
        except (UnicodeEncodeError, Exception):
            print('\n\nUsing these symbols:\n\n'
                  '  OK Optimized file      i EXIF info present\n'
                  '  -- Skipped file        V Image was downsized      v Size reduction')
            self.info = 'i'
            self.downsized = 'V '
            self.optimized = 'OK'
            self.skipped = '--'
            self.size_is_smaller = 'v'


def adjust_for_platform() -> Tuple[int, Union[TPoolExType, PPoolExType], int]:
    if platform.system() == 'Darwin':
        if platform.machine().startswith('iPad'):
            device = "iPad"
        elif platform.machine().startswith('iP'):
            device = "iPhone"
        else:
            device = "mac"
    else:
        device = "other"

    if device in ("iPad", "iPhone"):
        # Adapt for smaller screen sizes in iPhone and iPod touch
        import ui
        import console
        console.clear()
        if device == 'iPad':
            font_size = IPAD_FONT_SIZE
        else:
            font_size = IPHONE_FONT_SIZE
        console.set_font(IOS_FONT, font_size)
        screen_width = ui.get_screen_size().width
        char_width = ui.measure_string('.', font=(IOS_FONT, font_size)).width
        line_width = int(screen_width / char_width - 1.5) - 1
        pool_ex = concurrent.futures.ThreadPoolExecutor
        workers = IOS_WORKERS
    else:
        line_width = shutil.get_terminal_size((80, 24)).columns
        pool_ex = concurrent.futures.ProcessPoolExecutor
        from multiprocessing import cpu_count
        workers = cpu_count() + 1

    return line_width, pool_ex, workers



# encoding: utf-8

# ============================[ General settings ]============================
SUPPORTED_FORMATS = ['png', 'jpg', 'jpeg']
DEFAULT_QUALITY = 80
DEFAULT_BG_COLOR = (255, 255, 255)
MIN_BIG_IMG_SIZE = 80_000
MIN_BIG_IMG_AREA = 800 * 600

# ====================[ iOS/Pythonista specific settings ]====================
IPAD_FONT_SIZE = 15
IPHONE_FONT_SIZE = 10
IOS_WORKERS = 2
IOS_FONT = "Menlo"




# encoding: utf-8
import os

from typing import Iterable

from optimize_images.constants import SUPPORTED_FORMATS


def search_images(dirpath: str, recursive: bool) -> Iterable[str]:
    if recursive:
        for root, dirs, files in os.walk(dirpath):
            for f in files:
                if not os.path.isfile(os.path.join(root, f)):
                    continue
                extension = os.path.splitext(f)[1][1:]
                if extension.lower() in SUPPORTED_FORMATS:
                    yield os.path.join(root, f)
    else:
        with os.scandir(dirpath) as directory:
            for f in directory:
                if not os.path.isfile(os.path.normpath(f)):
                    continue
                extension = os.path.splitext(f)[1][1:]
                if extension.lower() in SUPPORTED_FORMATS:
                    yield os.path.normpath(f)



# encoding: utf-8
import os
import shutil

try:
    from PIL import Image, ImageFile
except ImportError:
    msg = 'This application requires Pillow to be installed. Please, install it first.'
    raise ImportError(msg)

from optimize_images.data_structures import Task, TaskResult
from optimize_images.img_info import is_big_png_photo
from optimize_images.img_aux_processing import remove_transparency, make_grayscale
from optimize_images.img_aux_processing import do_reduce_colors, downsize_img, rebuild_palette
from optimize_images.reporting import show_img_exception


def optimize_png(t: Task) -> TaskResult:
    """ Try to reduce file size of a PNG image.

        Expects a Task object containing all the parameters for the image processing.

        If file reduction is successful, this function will replace the original
        file with the optimized version and return some report data (file path,
        image format, image color mode, original file size, resulting file size,
        and resulting status of the optimization.

        :param t: A Task object containing all the parameters for the image processing.
        :return: A TaskResult object containing information for single file report.
        """
    img = Image.open(t.src_path)
    orig_format = img.format
    orig_mode = img.mode

    folder, filename = os.path.split(t.src_path)

    if folder == '':
        folder = os.getcwd()

    temp_file_path = os.path.join(folder + "/~temp~" + filename)
    orig_size = os.path.getsize(t.src_path)
    orig_colors, final_colors = 0, 0

    had_exif = has_exif = False  # Currently no exif methods for PNG files
    if orig_mode == 'P':
        final_colors = orig_colors = len(img.getcolors())

    if t.convert_all or (t.conv_big and is_big_png_photo(t.src_path)):
        # convert to jpg format
        filename = os.path.splitext(os.path.basename(t.src_path))[0]
        conv_file_path = os.path.join(folder + "/" + filename + ".jpg")

        if t.max_w or t.max_h:
            img, was_downsized = downsize_img(img, t.max_w, t.max_h)
        else:
            was_downsized = False

        img = remove_transparency(img, t.bg_color)
        img = img.convert("RGB")

        if t.grayscale:
            img = make_grayscale(img)

        try:
            img.save(
                conv_file_path,
                quality=t.quality,
                optimize=True,
                progressive=True,
                format="JPEG")
        except IOError:
            ImageFile.MAXBLOCK = img.size[0] * img.size[1]
            img.save(
                conv_file_path,
                quality=t.quality,
                optimize=True,
                progressive=True,
                format="JPEG")

        # Only save the converted file if conversion did save any space
        final_size = os.path.getsize(conv_file_path)
        if t.no_size_comparison or (orig_size - final_size > 0):
            was_optimized = True
            if t.force_del:
                try:
                    os.remove(t.src_path)
                except OSError as e:
                    details = 'Error while replacing original PNG with the new JPEG version.'
                    show_img_exception(e, t.src_path, details)
        else:
            final_size = orig_size
            was_optimized = False
            try:
                os.remove(conv_file_path)
            except OSError as e:
                details = 'Error while removing temporary JPEG converted file.'
                show_img_exception(e, t.src_path, details)

        result_format = "JPEG"
        return TaskResult(t.src_path, orig_format, result_format,
                          orig_mode, img.mode, orig_colors, final_colors,
                          orig_size, final_size, was_optimized,
                          was_downsized, had_exif, has_exif)

    # if PNG and user didn't ask for PNG to JPEG conversion, do this instead.
    else:
        result_format = "PNG"
        if t.remove_transparency:
            img = remove_transparency(img, t.bg_color)

        if t.max_w or t.max_h:
            img, was_downsized = downsize_img(img, t.max_w, t.max_h)
        else:
            was_downsized = False

        if t.reduce_colors:
            img, orig_colors, final_colors = do_reduce_colors(
                img, t.max_colors)

        if t.grayscale:
            img = make_grayscale(img)

        if not t.fast_mode and img.mode == "P":
            img, final_colors = rebuild_palette(img)

        try:
            img.save(temp_file_path, optimize=True, format=result_format)
        except IOError:
            ImageFile.MAXBLOCK = img.size[0] * img.size[1]
            img.save(temp_file_path, optimize=True, format=result_format)

        final_size = os.path.getsize(temp_file_path)

        # Only replace the original file if compression did save any space
        if t.no_size_comparison or (orig_size - final_size > 0):
            shutil.move(temp_file_path, os.path.expanduser(t.src_path))
            was_optimized = True
        else:
            final_size = orig_size
            was_optimized = False
            try:
                os.remove(temp_file_path)
            except OSError as e:
                details = 'Error while removing temporary file.'
                show_img_exception(e, t.src_path, details)

        return TaskResult(t.src_path, orig_format, result_format, orig_mode,
                          img.mode, orig_colors, final_colors, orig_size,
                          final_size, was_optimized, was_downsized, had_exif,
                          has_exif)



# encoding: utf-8

import concurrent.futures

from typing import NamedTuple, Tuple, NewType

try:
    from PIL import Image
except ImportError:
    msg = 'This application requires Pillow to be installed. Please, install it first.'
    raise ImportError(msg)

ImageType = NewType('ImageType', Image)
PPoolExType = NewType('PPoolExType', concurrent.futures.ProcessPoolExecutor)
TPoolExType = NewType('PPoolExType', concurrent.futures.ThreadPoolExecutor)


class Task(NamedTuple):
    src_path: str
    quality: int
    remove_transparency: bool
    reduce_colors: bool
    max_colors: int
    max_w: int
    max_h: int
    keep_exif: bool
    convert_all: bool
    conv_big: bool
    force_del: bool
    bg_color: Tuple[int, int, int]
    grayscale: bool
    no_size_comparison: bool
    fast_mode: bool


class TaskResult(NamedTuple):
    img: str
    orig_format: str
    result_format: str
    orig_mode: str
    result_mode: str
    orig_colors: int
    final_colors: int
    orig_size: int
    final_size: int
    was_optimized: bool
    was_downsized: bool
    had_exif: bool
    has_exif: bool



#!/usr/bin/env python3
# encoding: utf-8
"""
A little command-line interface (CLI) utility written in pure Python to help
you reduce the file size of images.

You must explicitly pass it a path to the source image file or to the
directory containing the image files to be processed. By default, it will go
through all of its subdirectories and try to optimize the images found. You
may however choose to process the specified directory only, without recursion.

Please note that the operation is done DESTRUCTIVELY, by replacing the
original files with the processed ones. You definitely should duplicate the
source file or folder before using this utility, in order to be able to
recover any eventual damaged files or any resulting images that don't have the
desired quality.

This application is intended to be pure Python, with no special dependencies
besides Pillow, therefore ensuring compatibility with a wide range of systems,
including iPhones and iPads running Pythonista 3. If you don't have the need
for such a strict dependency management, you will certainly be better served
by any several other image optimization utilities that are based on some well
known external binaries.

© 2020 Victor Domingos (MIT License)
"""
import concurrent.futures.process
import os

try:
    from PIL import Image
except ImportError:
    print('\n    This application requires Pillow to be installed. Please, install it first.\n')
    exit()

from timeit import default_timer as timer

from optimize_images.file_utils import search_images
from optimize_images.data_structures import Task
from optimize_images.do_optimization import do_optimization
from optimize_images.platforms import adjust_for_platform, IconGenerator
from optimize_images.argument_parser import get_args
from optimize_images.reporting import show_file_status, show_final_report
from optimize_images.reporting import show_img_exception


def main():
    appstart = timer()
    line_width, our_pool_executor, workers = adjust_for_platform()
    (src_path, recursive, quality, remove_transparency, reduce_colors,
     max_colors, max_w, max_h, keep_exif, convert_all, conv_big, force_del, bg_color,
     grayscale, ignore_size_comparison, fast_mode) = get_args()
    found_files = 0
    optimized_files = 0
    total_src_size = 0
    total_bytes_saved = 0

    icons = IconGenerator()

    # Optimize all images in a directory
    if os.path.isdir(src_path):
        recursion_txt = 'Recursively searching' if recursive else 'Searching'
        opt_msg = 'and optimizing image files'
        exif_txt = '(keeping exif data) ' if keep_exif else ''
        print(f"\n{recursion_txt} {opt_msg} {exif_txt}in:\n{src_path}\n")

        tasks = (Task(img_path, quality, remove_transparency, reduce_colors,
                      max_colors, max_w, max_h, keep_exif, convert_all, conv_big,
                      force_del, bg_color, grayscale, ignore_size_comparison, fast_mode)
                 for img_path in search_images(src_path, recursive=recursive)
                 if '~temp~' not in img_path)

        with our_pool_executor(max_workers=workers) as executor:
            current_img = ''
            try:
                for r in executor.map(do_optimization, tasks):
                    current_img = r.img
                    found_files += 1
                    total_src_size += r.orig_size
                    if r.was_optimized:
                        optimized_files += 1
                        total_bytes_saved += r.orig_size - r.final_size
                    show_file_status(r, line_width, icons)
            except concurrent.futures.process.BrokenProcessPool as e:
                show_img_exception(e, current_img)


    # Optimize a single image
    elif os.path.isfile(src_path) and '~temp~' not in src_path:
        found_files += 1

        img_task = Task(src_path, quality, remove_transparency, reduce_colors,
                        max_colors, max_w, max_h, keep_exif, convert_all, conv_big,
                        force_del, bg_color, grayscale, ignore_size_comparison, fast_mode)

        r = do_optimization(img_task)
        total_src_size = r.orig_size
        if r.was_optimized:
            optimized_files = 1
            total_bytes_saved = r.orig_size - r.final_size
        show_file_status(r, line_width, icons)
    else:
        print(
            "No image files were found. Please enter a valid path to the "
            "image file or the folder containing any images to be processed.")
        exit()

    if found_files:
        time_passed = timer() - appstart
        show_final_report(found_files, optimized_files, total_src_size,
                          total_bytes_saved, time_passed)
    else:
        print("No supported image files were found in the specified directory.\n")


if __name__ == "__main__":
    main()



# encoding: utf-8
import os

from io import BytesIO

try:
    from PIL import Image, ImageFile
except ImportError:
    msg = 'This application requires Pillow to be installed. Please, install it first.'
    raise ImportError(msg)

from optimize_images.img_aux_processing import downsize_img
from optimize_images.constants import MIN_BIG_IMG_SIZE, MIN_BIG_IMG_AREA


def is_big_png_photo(src_path: str) -> bool:
    """Try to determine if a given image if a big photo in PNG format

    Expects a path to a PNG image file. Returns True if the image is a PNG
    with an area bigger than MIN_BIG_IMG_AREA pixels that when resized to 1600
    pixels (wide or high) converts to a JPEG bigger than MIN_BIG_IMG_SIZE.
    Returns False otherwise.

    Inspired by an idea first presented by Stephen Arthur
    (https://engineeringblog.yelp.com/2017/06/making-photos-smaller.html)
    """
    img = Image.open(src_path)
    orig_format = img.format
    orig_mode = img.mode

    if orig_format != 'PNG' or orig_mode in ['P', 'L', 'LA']:
        return False

    w, h = img.size
    if (w * h) >= MIN_BIG_IMG_AREA:
        unique_colors = {img.getpixel((x, y)) for x in range(w) for y in range(h)}
        if len(unique_colors) > 2 ** 16:
            img = img.convert("RGB")
            if w > h:
                img, status = downsize_img(img, 1600, 0)
            else:
                img, status = downsize_img(img, 0, 1600)

            tempfile = BytesIO()
            try:
                img.save(tempfile, quality=80, format="JPEG")
            except IOError:
                ImageFile.MAXBLOCK = img.size[0] * img.size[1]
                img.save(tempfile, quality=80, format="JPEG")

            final_size = tempfile.getbuffer().nbytes
            return final_size > MIN_BIG_IMG_SIZE

    return False



# encoding: utf-8
from typing import Tuple

try:
    from PIL import Image
except ImportError:
    msg = 'This application requires Pillow to be installed. Please, install it first.'
    raise ImportError(msg)

from optimize_images.constants import DEFAULT_BG_COLOR
from optimize_images.data_structures import ImageType


class Palette:
    def __init__(self):
        self.palette = []

    def add(self, r, g, b):
        # map rgb tuple to colour index
        rgb = r, g, b
        try:
            return self.palette.index(rgb)
        except:
            i = len(self.palette)
            if i >= 256:
                raise RuntimeError("all palette entries are used")
            self.palette.append(rgb)
            return i

    def get_palette(self):
        # return flattened palette
        palette = []
        for r, g, b in self.palette:
            palette = palette + [r, g, b]
        return palette


def remove_transparency(img: ImageType, bg_color=DEFAULT_BG_COLOR) -> ImageType:
    """Remove alpha transparency from PNG images

    Expects a PIL.Image object and returns an object of the same type with the
    changes applied.

    Special thanks to Yuji Tomita and Takahashi Shuuji
    (https://stackoverflow.com/a/33507138)
    """
    if img.mode in ('RGBA', 'LA') or (img.mode == 'P' and 'transparency' in img.info):
        orig_image = img.convert('RGBA')
        background = Image.new('RGBA', orig_image.size, bg_color)
        img = Image.alpha_composite(background, orig_image)
        return img.convert("RGB")
    else:
        return img


def downsize_img(img: ImageType, max_w: int, max_h: int) -> Tuple[ImageType, bool]:
    """ Reduce the size of an image to the indicated maximum dimensions

    This function takes a PIL.Image object and integer values for the maximum
    allowed width and height (a zero value means no maximum constraint),
    calculates the size that meets those constraints and resizes the image. The
    resize is done in place, changing the original object. Returns a boolean
    indicating if the image was changed.
    """
    w, h = img.size
    # Assume 0 as current size
    if not max_w:
        max_w = w
    if not max_h:
        max_h = h

    if (max_w, max_h) == (w, h):  # If no changes, do nothing
        return img, False

    img.thumbnail((max_w, max_h), resample=Image.LANCZOS)
    return img, True


def do_reduce_colors(img: ImageType, max_colors: int) -> Tuple[ImageType, int, int]:
    """ Reduce the number of colors of an Image object

    It takes a PIL image object and tries to reduce the total number of colors,
    converting it to an indexed color (mode P) image. If the input image is in
    mode 1, it cannot be further reduced, so it's returned back with no
    changes.

    :param img: a PIL image in color (modes P, RGBA, RGB, CMYK, YCbCr, LAB or HSV)
    :param max_colors: an integer indicating the maximum number of colors allowed.
    :return: a PIL image in mode P (or mode 1, as stated above), an integer
             indicating the original number of colors (0 if source is not a
             mode P or mode 1 image) and an integer stating the resulting
             number of colors.
    """
    orig_mode = img.mode

    if orig_mode == "1":
        return img, 2, 2

    colors = img.getcolors()
    if colors:
        orig_colors = len(colors)
    else:
        orig_colors = 0

    # Intermediate conversion steps when needed
    if orig_mode in ["CMYK", "YCbCr", "LAB", "HSV"]:
        img = img.convert("RGB")
    elif orig_mode == "LA":
        img = img.convert("RGBA")

    # Actual color reduction happening here
    if orig_mode in ["RGB", "L"]:
        palette = Image.ADAPTIVE
    elif orig_mode == "RGBA":
        palette = Image.ADAPTIVE
        transparent = Image.new("RGBA", img.size, (0, 0, 0, 0))
        # blend with transparent image using own alpha
        img = Image.composite(img, transparent, img)
    elif orig_mode == "P":
        palette = img.getpalette()
        img = img.convert("RGBA")
        w, h = img.size
        alpha_layer = Image.new("L", img.size)
        for x in range(w):
            for y in range(h):
                r, g, b, a = img.getpixel((x, y))
                alpha_layer.putpixel((x, y), a)
        img.putalpha(alpha_layer)
    else:
        return img, 0, 0

    img = img.convert("P", palette=palette, colors=max_colors)
    return img, orig_colors, len(img.getcolors())


def make_grayscale(img: ImageType) -> ImageType:
    """ Convert an Image to grayscale

    :param img: a PIL image in color (modes P, RGBA, RGB, CMYK, YCbCr, LAB or HSV)
    :return: a PIL image object in modes P, L or RGBA, if converted, or the
             original Image object in case no conversion is done.
    """
    orig_mode = img.mode

    if orig_mode in ["RGB", "CMYK", "YCbCr", "LAB", "HSV"]:
        return img.convert("L")
    elif orig_mode == "RGBA":
        return img.convert("LA").convert("RGBA")
    elif orig_mode == "P":
        # Using ITU-R 601-2 luma transform:  L = R * 299/1000 + G * 587/1000 + B * 114/1000
        pal = img.getpalette()
        for i in range(len(pal) // 3):
            # Using ITU-R 601-2 luma transform
            g = (pal[3 * i] * 299 + pal[3 * i + 1] * 587 + pal[3 * i + 2] * 114) // 1000
            pal[3 * i: 3 * i + 3] = [g, g, g]
        img.putpalette(pal)
        return img
    else:
        return img


def rebuild_palette(img: ImageType) -> Tuple[ImageType, int]:
    """ Rebuild the palette of a mode "P" PNG image

    It may allow for other tools, like PNGOUT and AdvPNG, to further reduce the
    size of some indexed PNG images. However, it it was already an optimized PNG,
    the resulting file size may in fact be bigger (which means optimize-images
    may discard it by default). You may also want to use it as an intermediate
    process, before doing a final optimization using those tools.

    :param img: a mode "P" PNG image
    :return: a tuple composed by a mode "P" PNG image object and an integer
             with the resulting number of colors
    """
    w, h = img.size
    img = img.convert("RGBA")
    new_palette = Palette()
    alpha_layer = Image.new("L", img.size)

    for x in range(w):
        for y in range(h):
            r, g, b, a = img.getpixel((x, y))
            alpha_layer.putpixel((x, y), a)
            new_palette.add(r, g, b)

    img.putalpha(alpha_layer)
    palette = new_palette.get_palette()
    num_colors = len(palette) // 3
    img = img.convert("P", palette=palette, colors=num_colors)
    return img, len(img.getcolors())



# encoding: utf-8
from optimize_images.data_structures import TaskResult
from optimize_images.platforms import IconGenerator


def human(number: int, suffix='B') -> str:
    """Return a human readable memory size in a string.

    Initially written by Fred Cirera, modified and shared by Sridhar Ratnakumar
    (https://stackoverflow.com/a/1094933/6167478), edited by Victor Domingos.
    """
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(number) < 1024.0:
            return f"{number:3.1f} {unit}{suffix}"
        number = number / 1024.0
    return f"{number:.1f}{'Yi'}{suffix}"


def show_file_status(r: TaskResult, line_width: int, icons: IconGenerator):
    if r.was_optimized:
        short_img = r.img[-(line_width - 17):].ljust(line_width - 17)
        percent = 100 - (r.final_size / r.orig_size * 100)
        h_orig = human(r.orig_size)
        h_final = human(r.final_size)

        orig_format = r.orig_format.replace('JPEG', 'JPG')
        result_format = r.result_format.replace('JPEG', 'JPG')

        if r.orig_mode == "P":
            o_colors = f"{r.orig_colors}"
        else:
            o_colors = ""

        if r.result_mode == "P":
            colors = f"{r.final_colors}"
        else:
            colors = ""

        exif_str1 = icons.info if r.had_exif else ''
        exif_str2 = icons.info if r.has_exif else ''
        downstr = icons.downsized if r.was_downsized else ''
        line1 = f'\n{icons.optimized}  [OPTIMIZED] {short_img}\n'
        line2 = f'    {exif_str1} {orig_format}/{r.orig_mode}{o_colors}: {h_orig}  ->  {downstr}{exif_str2}{result_format}/{r.result_mode}{colors}: {h_final} {icons.size_is_smaller} {percent:.1f}%'
        img_status = line1 + line2
    else:
        short_img = r.img[-(line_width - 15):].ljust(line_width - 15)
        img_status = f'\n{icons.skipped}  [SKIPPED] {short_img}'

    print(img_status, end='')


def show_final_report(found_files: int,
                      optimized_files: int,
                      src_size: int,
                      bytes_saved: int,
                      time_passed: float):
    fps = found_files / time_passed

    if bytes_saved:
        average = bytes_saved / optimized_files
        percent = bytes_saved / src_size * 100
    else:
        average = 0
        percent = 0

    print(f"\n{40 * '-'}\n")
    print(
        f"   Processed {found_files} files ({human(src_size)}) in {time_passed:.1f}s ({fps:.1f} f/s)."
    )
    print(f"   Optimized {optimized_files} files.")
    print(f"   Average savings: {human(average)} per optimized file")
    print(f"   Total space saved: {human(bytes_saved)} / {percent:.1f}%\n")


def show_img_exception(exception: Exception, image_path: str, details: str = '') -> None:
    print("\nAn error has occurred while trying to optimize this file:")
    print(image_path)

    if details:
        print(f'\n{details}')

    print("\nThe following info may help to understand what has gone wrong here:\n")
    print(exception)



"""
Adapted from:
https://engineeringblog.yelp.com/2017/06/making-photos-smaller.html
"""
from io import BytesIO
from math import log

from optimize_images.constants import DEFAULT_QUALITY

try:
    from PIL import Image
    from PIL import ImageChops, ImageStat
except ImportError:
    msg = 'This application requires Pillow to be installed. Please, install it first.'
    raise ImportError(msg)


def compare_images(img1, img2):
    """Calculate the difference between two images of the same size
    by comparing channel values at the pixel level.
    `delete_diff_file`: removes the diff image after ratio found
    `diff_img_file`: filename to store diff image

    Adapted from Nicolas Hahn:
    https://github.com/nicolashahn/diffimg/blob/master/diffimg/__init__.py
    """

    # Don't compare if images are of different modes or different sizes.
    if (img1.mode != img2.mode) \
            or (img1.size != img2.size) \
            or (img1.getbands() != img2.getbands()):
        return None

    # Generate diff image in memory.
    diff_img = ImageChops.difference(img1, img2)
    # Calculate difference as a ratio.
    stat = ImageStat.Stat(diff_img)
    diff_ratio = sum(stat.mean) / (len(stat.mean) * 255)

    return diff_ratio * 100


def get_diff_at_quality(photo, quality):
    """Return a difference score for this JPEG image saved at the specified quality

    A SSIM score would be much better, but currently there is no pure Python
    implementation available.
    """
    diff_photo = BytesIO()
    # optimize is omitted here as it doesn't affect
    # quality but requires additional memory and cpu
    photo.save(diff_photo, format="JPEG", quality=quality, progressive=True)
    diff_photo.seek(0)
    diff_score = compare_images(photo, Image.open(diff_photo))

    # print("================> DIFF1 == DIFF2? ", diff_score==diff_score2)

    if diff_score < 0:
        return -1 + diff_score / 100
    else:
        return 1 - diff_score / 100


def _diff_iteration_count(lo, hi):
    """Return the depth of the binary search tree for this range"""
    if lo >= hi:
        return 0
    else:
        return int(log(hi - lo, 2)) + 1


def jpeg_dynamic_quality(original_photo, use_dynamic_quality=True):
    """Return an integer representing the quality that this JPEG image should be
    saved at to attain the quality threshold specified for this photo class.

    Args:
        original_photo - a prepared PIL JPEG image (only JPEG is supported)
    """
    diff_goal = 0.992
    hi = DEFAULT_QUALITY
    lo = hi - 5

    # working on a smaller size image doesn't give worse results but is faster
    # changing this value requires updating the calculated thresholds
    photo = original_photo.resize((400, 400))

    if not use_dynamic_quality:
        default_diff = get_diff_at_quality(photo, hi)
        return hi, default_diff

    # 95 is the highest useful value for JPEG. Higher values cause different behavior
    # Used to establish the image's intrinsic ssim without encoder artifacts
    normalized_diff = get_diff_at_quality(photo, 95)

    selected_quality = selected_diff = None

    # loop bisection. ssim/diff function increases monotonically so this will converge
    for i in range(_diff_iteration_count(lo, hi)):
        curr_quality = (lo + hi) // 2
        curr_diff = get_diff_at_quality(photo, curr_quality)
        diff_ratio = curr_diff / normalized_diff

        if diff_ratio >= diff_goal:
            # continue to check whether a lower quality level also exceeds the goal
            selected_quality = curr_quality
            selected_diff = curr_diff
            hi = curr_quality
        else:
            lo = curr_quality

    if selected_quality:
        return selected_quality, selected_diff
    else:
        default_diff = get_diff_at_quality(photo, hi)
        return hi, default_diff



# encoding: utf-8
import os
import shutil

import piexif

try:
    from PIL import Image, ImageFile
except ImportError:
    msg = 'This application requires Pillow to be installed. Please, install it first.'
    raise ImportError(msg)

from optimize_images.data_structures import Task, TaskResult
from optimize_images.img_aux_processing import make_grayscale
from optimize_images.img_aux_processing import downsize_img
from optimize_images.img_dynamic_quality import jpeg_dynamic_quality
from optimize_images.reporting import show_img_exception


def optimize_jpg(t: Task) -> TaskResult:
    """ Try to reduce file size of a JPG image.

    Expects a Task object containing all the parameters for the image processing.

    If file reduction is successful, this function will replace the original
    file with the optimized version and return some report data (file path,
    image format, image color mode, original file size, resulting file size,
    and resulting status of the optimization.

    :param t: A Task object containing all the parameters for the image processing.
    :return: A TaskResult object containing information for single file report.
    """
    img = Image.open(t.src_path)
    orig_format = img.format
    orig_mode = img.mode

    folder, filename = os.path.split(t.src_path)

    if folder == '':
        folder = os.getcwd()

    temp_file_path = os.path.join(folder + "/~temp~" + filename)
    orig_size = os.path.getsize(t.src_path)
    orig_colors, final_colors = 0, 0

    result_format = "JPEG"
    try:
        had_exif = True if piexif.load(t.src_path)['Exif'] else False
    except piexif.InvalidImageDataError:  # Not a supported format
        had_exif = False
    except ValueError:  # No exif info
        had_exif = False
    # TODO: Check if we can provide a more specific treatment of piexif exceptions.
    except Exception:
        had_exif = False

    if t.max_w or t.max_h:
        img, was_downsized = downsize_img(img, t.max_w, t.max_h)
    else:
        was_downsized = False

    if t.grayscale:
        img = make_grayscale(img)

    # only use progressive if file size is bigger
    use_progressive_jpg = orig_size > 10000

    if t.fast_mode:
        quality = t.quality
    else:
        quality, jpgdiff = jpeg_dynamic_quality(img)

    try:
        img.save(
            temp_file_path,
            quality=quality,
            optimize=True,
            progressive=use_progressive_jpg,
            format=result_format)
    except IOError:
        ImageFile.MAXBLOCK = img.size[0] * img.size[1]
        img.save(
            temp_file_path,
            quality=quality,
            optimize=True,
            progressive=use_progressive_jpg,
            format=result_format)

    if t.keep_exif and had_exif:
        try:
            piexif.transplant(
                os.path.expanduser(t.src_path), temp_file_path)
            has_exif = True
        except ValueError:
            has_exif = False
        # TODO: Check if we can provide a more specific treatment of piexif exceptions.
        except Exception:
            had_exif = False
    else:
        has_exif = False

    # Only replace the original file if compression did save any space
    final_size = os.path.getsize(temp_file_path)
    if t.no_size_comparison or (orig_size - final_size > 0):
        shutil.move(temp_file_path, os.path.expanduser(t.src_path))
        was_optimized = True
    else:
        final_size = orig_size
        was_optimized = False
        try:
            os.remove(temp_file_path)
        except OSError as e:
            details = 'Error while removing temporary file.'
            show_img_exception(e, t.src_path, details)

    return TaskResult(t.src_path, orig_format, result_format, orig_mode,
                      img.mode, orig_colors, final_colors, orig_size,
                      final_size, was_optimized, was_downsized, had_exif,
                      has_exif)



# encoding: utf-8
from setuptools import setup, find_packages
import os
import sys
import platform

used = sys.version_info
required = (3, 6)

# if version of pip that doesn't understand the python_requires classifier, must be pip >= 9.0.0
# must be built using at least version 24.2.0 of setuptools
# in order for the python_requires argument to be recognized and the appropriate metadata generated
# python -m pip install --upgrade pip setuptools
if used[:2] < required:
    sys.stderr.write("Unsupported Python version: %s.%s. "
                     "Python 3.6 or later is required." % (sys.version_info.major, sys.version_info.minor))
    sys.exit(1)

short_desc = "A command-line interface (CLI) utility written in pure Python " \
             "to help you reduce the file size of images."


def read_readme(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name)) as f:
        return f.read()


setup(name='optimize-images',
      version=__import__('optimize_images').__version__,
      description=short_desc,
      author="Victor Domingos",
      packages=find_packages(),
      include_package_data=False,
      long_description=read_readme('README.md'),  # for PyPI
      long_description_content_type="text/markdown",
      license='MIT',
      url='https://no-title.victordomingos.com/projects/optimize-images/',
      project_urls={
          'Documentation': 'https://github.com/victordomingos/optimize-images/blob/master/docs/docs_EN.md',
          'Source': 'https://github.com/victordomingos/optimize-images',
          'Bug Reports': 'https://github.com/victordomingos/optimize-images/issues',
      },
      python_requires='>=3.6',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Operating System :: iOS',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: Unix',
          'Operating System :: POSIX :: Linux ',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Utilities',
          'Topic :: Multimedia :: Graphics',
          'Topic :: Multimedia :: Graphics :: Graphics Conversion',
      ],

      keywords='python3 pythonista-ios pil pillow image-processing ' \
               'image-compression image-optimization image-optimisation seo '
               'seo-optimization website-performance cli recursive non-recursive',

      install_requires=[
          'piexif>=1.1.3',
      ],

      entry_points={
          'console_scripts': ['optimize-images = optimize_images.__main__:main']
      },
      )



# encoding: utf-8

try:
    from PIL import Image
except ImportError:
    print('\n    This application requires Pillow to be installed. Please, install it first.\n')
    exit()
#
# try:
#     from PIL import UnidentifiedImageError
# except ImportError:
#     print('\n    This application requires Pillow to be installed. Please, install it first.aaa\n')
#     exit()

import os
import piexif

from optimize_images.data_structures import Task, TaskResult
from optimize_images.img_optimize_jpg import optimize_jpg
from optimize_images.img_optimize_png import optimize_png


def do_optimization(t: Task) -> TaskResult:
    """ Try to reduce file size of an image.

    Expects a Task object containing all the parameters for the image processing.

    The actual processing is done by the corresponding function,
    according to the detected image format.

    :param t: A Task object containing all the parameters for the image processing.
    :return: A TaskResult object containing information for single file report.
    """
    # TODO: Catch exceptions that may occur here.
    try:
        img = Image.open(t.src_path)

        # TODO: improve method of image format detection (what should happen if the
        #       file extension does not match the image content's format? Maybe we
        #       should skip unsupported formats?)
        if img.format.upper() == 'PNG':
            return optimize_png(t)
        elif img.format.upper() in ('JPEG', 'MPO'):
            return optimize_jpg(t)
        else:
            try:
                had_exif = True if piexif.load(t.src_path)['Exif'] else False
            except piexif.InvalidImageDataError:  # Not a supported format
                had_exif = False
            except ValueError:  # No exif info
                had_exif = False
            return TaskResult(img=t.src_path,
                              orig_format=img.format.upper(),
                              result_format=img.format.upper(),
                              orig_mode=img.mode,
                              result_mode=img.mode,
                              orig_colors=0,
                              final_colors=0,
                              orig_size=os.path.getsize(t.src_path),
                              final_size=0,
                              was_optimized=False,
                              was_downsized=False,
                              had_exif=had_exif,
                              has_exif=had_exif)

    except IOError:
        try:
            had_exif = True if piexif.load(t.src_path)['Exif'] else False
        except piexif.InvalidImageDataError:  # Not a supported format
            had_exif = False
        except ValueError:  # No exif info
            had_exif = False
        return TaskResult(img=t.src_path,
                          orig_format="",
                          result_format="",
                          orig_mode="",
                          result_mode="",
                          orig_colors=0,
                          final_colors=0,
                          orig_size=os.path.getsize(t.src_path),
                          final_size=0,
                          was_optimized=False,
                          was_downsized=False,
                          had_exif=had_exif,
                          has_exif=had_exif)
